use crate::{math::{PixelPoint, PixelBox, PixelSize}, layout::{SizePolicy, LayoutNodeAccess, cache::NodeState}};
use super::{ArrangementCalculator, do_fit};
use std::ops::Add;

#[derive(Clone, Debug)]
pub enum LineArrangement {
    Column,
    Row,
}

impl LineArrangement {
    fn is_column(&self) -> bool {
        match self {
            Self::Column => true,
            Self::Row => false,
        }
    }
}

impl ArrangementCalculator for LineArrangement {
    fn arrange_step(
        &self,
        node: LayoutNodeAccess,
        child_policies: Vec<(SizePolicy, SizePolicy)>,
    ) -> (SizePolicy, SizePolicy) {
        if child_policies.is_empty() {
            return (node.width_policy, node.height_policy);
        }

        let cw_policy = child_policies
            .iter()
            .map(|v| v.0)
            .reduce(if self.is_column() {
                SizePolicy::max
            } else {
                SizePolicy::add
            })
            .unwrap();
        let ch_policy: SizePolicy = child_policies
            .iter()
            .map(|v| v.1)
            .reduce(if self.is_column() {
                SizePolicy::add
            } else {
                SizePolicy::max
            })
            .unwrap();
        (
            node.width_policy.max_preserve_slack(cw_policy),
            node.height_policy.max_preserve_slack(ch_policy),
        )
    }

    fn layout_step(&self, node: LayoutNodeAccess, inside: PixelBox) {
        // do the final children layouts
        node.cache
            .with_state(node.cache_key, |ns| {
                if Some(inside) != ns.area {
                    ns.area = Some(inside);
                    ns.needs_render = true;
                }
            });

        if node.child_len() == 0 {
            return;
        }

        // expansion direction extraction lambda
        let ee = if self.is_column() {
            |ns: &mut NodeState| ns.net_policy.1
        } else {
            |ns: &mut NodeState| ns.net_policy.0
        };

        let policies = node
            .child_iter()
            .map(|c| node.cache.with_state(c.cache_key, ee).unwrap());
        let mut last_offset = 0;
        let fit = do_fit(
            if self.is_column() {
                inside.height() as i32
            } else {
                inside.width() as i32
            },
            policies,
        );

        for (offset, child) in fit.zip(node.child_iter()) {
            let cbox = if self.is_column() {
                PixelBox::from_origin_and_size(
                    PixelPoint {
                        x: inside.min.x,
                        y: inside.min.y + last_offset,
                        ..Default::default()
                    },
                    PixelSize {
                        width: inside.width(),
                        height: (offset - last_offset) as i32,
                        ..Default::default()
                    },
                )
            } else {
                PixelBox::from_origin_and_size(
                    PixelPoint {
                        x: inside.min.x + last_offset,
                        y: inside.min.y,
                        ..Default::default()
                    },
                    PixelSize {
                        width: (offset - last_offset) as i32,
                        height: inside.height(),
                        ..Default::default()
                    },
                )
            };

            self.layout_step(child, cbox);
            last_offset = offset;
        }
    }
}
