use crate::{
    component::Component,
    input::InputState,
    layout::{LayoutNode, LayoutNodeAccess, LayoutNodeContainer, SizePolicy, ChildArrangement},
    theme::Theme,
    ui::UIHandle,
    widget::Widget, window::RenderFormat,
};

pub struct PlainGroup<C: Component> {
    lnode: LayoutNode,
    children: Vec<Box<dyn Widget<C>>>,
}

impl<C: Component> PlainGroup<C> {
    pub fn new(uih: &UIHandle) -> Self {
        Self {
            lnode: {
                let mut node = uih.new_layout_node();
                node.set_width_policy(SizePolicy::expanding(1));
                node.set_height_policy(SizePolicy::expanding(1));
                node
            },
            children: vec![],
        }
    }

    pub fn new_column(uih: &UIHandle) -> Self {
        let mut ret = Self::new(uih);
        ret.lnode.set_arrangement(ChildArrangement::Column);
        ret
    }

    pub fn new_row(uih: &UIHandle) -> Self {
        let mut ret = Self::new(uih);
        ret.lnode.set_arrangement(ChildArrangement::Row);
        ret
    }

    pub fn new_table(uih: &UIHandle) -> Self {
        let mut ret = Self::new(uih);
        ret.lnode.set_arrangement(ChildArrangement::Table);
        ret
    }

    pub fn append(&mut self, widget: Box<dyn Widget<C>>) {
        self.children.push(widget);
    }

    pub fn extend(&mut self, with: impl Iterator<Item = Box<dyn Widget<C>>>) {
        self.children.extend(with);
    }
}

impl<C: Component> LayoutNodeContainer for PlainGroup<C> {
    fn layout_node(&self) -> &LayoutNode {
        &self.lnode
    }

    fn layout_child(&self, ndx: usize) -> Option<LayoutNodeAccess> {
        self.children.get(ndx).map(|w| w.layout_node())
    }

    fn layout_child_count(&self) -> usize {
        self.children.len()
    }
}

impl<C: Component> Widget<C> for PlainGroup<C> {
    fn layout_node(&self) -> LayoutNodeAccess {
        LayoutNodeAccess::new(self)
    }
    fn layout_node_mut(&mut self) -> &mut LayoutNode {
        &mut self.lnode
    }
    fn poll(&mut self, uih: &mut UIHandle, input_state: Option<&InputState>) -> Vec<C::Msg> {
        self.children
            .iter_mut()
            .map(|w| w.poll(uih, input_state))
            .flatten()
            .collect()
    }
    fn render(&self, theme: &Theme, target: &mut kahlo::BitmapMut<RenderFormat>) {
        for child in self.children.iter() {
            child.render(theme, target);
        }
    }
}

use super::Label;

pub struct FormGroup<C: Component> {
    lnode: LayoutNode,
    labelnode: LayoutNode,
    widgetnode: LayoutNode,
    labels: Vec<Label<C>>,
    widgets: Vec<Box<dyn Widget<C>>>,
}

struct LabelContainer<'l, C: Component>(&'l FormGroup<C>);
struct WidgetContainer<'l, C: Component>(&'l FormGroup<C>);

impl<C: Component> LayoutNodeContainer for FormGroup<C> {
    fn layout_node(&self) -> &LayoutNode {
        &self.lnode
    }
    fn layout_child(&self, ndx: usize) -> Option<LayoutNodeAccess<'_>> {
        match ndx {
            0 => todo!(), // Some(LayoutNodeAccess::new(LabelContainer(self))),
            1 => todo!(),
            _ => None
        }
    }
    fn layout_child_count(&self) -> usize {
        2
    }
}

impl<C: Component> Widget<C> for FormGroup<C> {
    fn layout_node(&self) -> LayoutNodeAccess {
        todo!()
    }
    fn layout_node_mut(&mut self) -> &mut LayoutNode {
        todo!()
    }
    fn poll(&mut self, uih: &mut UIHandle, input_state: Option<&InputState>) -> Vec<<C as Component>::Msg> {
        todo!()
    }
    fn render(&self, theme: &Theme, target: &mut kahlo::BitmapMut<RenderFormat>) {
        
    }
}
